# Copyright (c) 2012 Rick Harris
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import collections


def settle(txns):
    """Return minimum set of transactions to cancel a given debt.

    :param txns: a set of transactions made up of tuples of form
                 (txn_id, debtor, creditor, amount)
    :returns: a set of transactions which cancel the debts incurred by input
              transactions
    """
    if not txns:
        return set()

    # To be type-agnostic, use whatever type was passed in as first element's
    # amount
    first_rec = next(iter(txns))
    type_to_use = type(first_rec[3])

    # Use hashes to pick which node goes first, A or B. Then the tuple (A, B)
    # reflects the edge and the amount reflects the weight.
    balances = collections.defaultdict(type_to_use)

    # These are the txn_ids that led to a given debt
    txn_ids_for_debt = collections.defaultdict(set)

    for txn_id, debtor, creditor, amount in txns:
        debtor_hash = hash(debtor)
        creditor_hash = hash(creditor)

        if debtor_hash > creditor_hash:
            balances[(debtor, creditor)] += amount
            txn_ids_for_debt[(debtor, creditor)].add(txn_id)
        elif debtor_hash < creditor_hash:
            balances[(creditor, debtor)] -= amount
            txn_ids_for_debt[(creditor, debtor)].add(txn_id)
        else:
            continue  # cannot owe yourself

    # Flip any negative amounts, so if A owes B -$10, then B owes A $10
    settlement = set()
    for (debtor, creditor), amount in balances.items():
        if not amount:
            continue

        txn_ids = frozenset(txn_ids_for_debt[(debtor, creditor)])
        if amount < 0:
            debtor, creditor = creditor, debtor
            amount = -amount

        settlement.add((debtor, creditor, amount, txn_ids))

    return settlement


def quantized_split(amount, n, quanta=0.01):
    """Split an amount n-ways with extra pennies from uneven splits placed at
    the front of the list.

    Example:

        split(10, 3) -> [3.34, 3.33, 3.33]

    :param amount: amount to split
    :param n: number of ways to split `amount`
    :returns: list containing split amount with heavier amounts first.
    """
    if n < 1:
        raise ValueError('n must be an integer >= 1')

    even, remainder = divmod(int(amount / quanta), n)

    heavy = int(remainder)
    light = n - heavy
    return ([even * quanta + quanta] * heavy) + ([even * quanta] * light)
